/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * Peak.cpp
 *
 *  Created on: 2010/12/13
 *      Author: HARK Development Team
 */

#include <Peak.hpp>
#include <LRSignal.hpp>
#include <Band.hpp>
#include <Logger.hpp>

using namespace std;
using namespace boost;

namespace NSHarkBinaural
{

Peak::LRPeaksPtr
Peak::GetPeaks(
    const LRUnspectra::ptr_t bgnoise,
    const LRUnspectra::ptr_t lrpower,
    const Band& lower,
    const Band& upper,
    const BandInformation& bi,
    //Unspectrum::ElementType minDB,
    Unspectrum::ElementType thresDB)
{
    static Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);
    // エラーチェック
    if (upper.Index() <= 0 || lower.Index() >= upper.Index()) // lower, upper自体の検査
    {
        LogWarning(logger, "upper or lower is invalid");
        return LRPeaksPtr();
    }
    else if (!IsEqualSignalSize(bgnoise) || !IsEqualSignalSize(lrpower)) // bgnoise, lrpower自体の検査
    {
        LogWarning(logger, "bgnoise or lrpower is invalid");
        return LRPeaksPtr();
    }
    else if (bgnoise->Left()->Size() < upper.Index() || lrpower->Left()->Size() < upper.Index()) // 少なくともupper.Index()個の要素があることの検査
    {
        LogWarning(logger, "size of bgnoise or lrpower is invalid");
        return LRPeaksPtr();
    }

    // 戻り値の生成
    LogFine(logger, format("creating return variables"));
    //Unspectrum::ptr_t peakL(new Unspectrum(upper.Index()));
    //Unspectrum::ptr_t peakR(new Unspectrum(upper.Index()));
    PeakVectorPtr peakL(new PeakVector());
    PeakVectorPtr peakR(new PeakVector());
    /*
    if (peakL.get() == 0 || peakR.get() == 0)
    {
      return LRUnspectra::ptr_t();
    }
    */
    //LRUnspectra::ptr_t peak(new LRUnspectra(peakL, peakR));
    LRPeaksPtr peak(new LRPeaks(peakL, peakR));

    // 使いやすいように参照を定義
    //Unspectrum& pkL = *peakL;
    //Unspectrum& pkR = *peakR;
    PeakVector& pkL = *peakL;
    PeakVector& pkR = *peakR;
    const Unspectrum& bgnL = *bgnoise->Left();
    const Unspectrum& bgnR = *bgnoise->Right();
    const Unspectrum& powL = *lrpower->Left();
    const Unspectrum& powR = *lrpower->Right();

    LogFine(logger, format("calculating peaks"));
    size_t i;

#ifndef NOCLUSTER
    /*
    for (i = 0; i < lower.Index(); ++i)
    {
      pkL[i] = pkR[i] = minDB;
    }
    for (; i < pkL.Size(); ++i)
    {
      if (powL.At(i) > bgnL.At(i) + thresDB && powR.At(i) > bgnR.At(i) + thresDB)
      {
        pkL[i] = powL[i];
        pkR[i] = powR[i];
      }
      else
      {
        pkL[i] = pkR[i] = minDB;
      }
    }
    */
    for (i = lower.Index(); i < powL.Size() && i < upper.Index(); ++i)
    {
        if (powL.At(i) > bgnL.At(i) + thresDB && powR.At(i) > bgnR.At(i) + thresDB)
        {
            pkL.push_back(BandAndPower(bi.CreateBand(i), powL[i]));
            pkR.push_back(BandAndPower(bi.CreateBand(i), powR[i]));
        }
    }
#else
    /*
    for (i = 0; i < lower.Index(); ++i)
    {
      pkL[i] = pkR[i] = minDB;
    }
    for (; i < pkL.Size() - 1; ++i)
    {
      if ((powL.At(i) > powL.At(i - 1) && powL.At(i) > powL.At(i + 1)) &&
          (powR.At(i) > powR.At(i - 1) && powR.At(i) > powR.At(i + 1)) &&
          (powL.At(i) > bgnL.At(i) + thresDB || powR.At(i) > bgnR.At(i) + thresDB))
      {
        pkL[i] = powL[i];
        pkR[i] = powR[i];
      }
      else
      {
        pkL[i] = pkR[i] = minDB;
      }
    }
    pkL[i] = pkR[i] = minDB;
    */
    for (i = lower.Index(); i < powL.Size() - 1 && i < upper.Index(); ++i)
    {
        if ((powL.At(i) > powL.At(i - 1) && powL.At(i) > powL.At(i + 1)) &&
                (powR.At(i) > powR.At(i - 1) && powR.At(i) > powR.At(i + 1)) &&
                (powL.At(i) > bgnL.At(i) + thresDB || powR.At(i) > bgnR.At(i) + thresDB))
        {
            pkL.push_back(BandAndPower(bi.CreateBand(i), powL[i]));
            pkR.push_back(BandAndPower(bi.CreateBand(i), powR[i]));
        }
    }
#endif
    return peak;
}

}
