/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 *
 */

#ifdef NDEBUG
#undef NDEBUG
#endif

#include <iostream>
#include <string>
#include <typeinfo>
#include <complex>

#include <boost/shared_ptr.hpp>
#include <boost/format.hpp>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Map.h>
#include <Source.h>
#include <libharkio/libharkio.h>

#include <Signal.hpp>
#include <Channel.hpp>
#include <InterauralDifference.hpp>
#include <DirectionPassFilter.hpp>
#include <Epipolar.hpp>
#include <Separation.hpp>
#include <LRSignal.hpp>
#include <InvalidInputException.hpp>
#include <Logger.hpp>
#include <HRTF.hpp>
#include <BGNEstimator.hpp>

#include <DimensionPromise/dimpro.hpp>

using namespace std;
using namespace boost;
using namespace FD;
using namespace NSHarkBinaural;
using namespace NSDimensionPromise;

class SeparateBinaural;

DECLARE_NODE(SeparateBinaural);
/*Node
 *
 * @name SeparateBinaural
 * @category HARK:BINAURAL
 * @description Separate sound sources
 *
 * @input_name SPECTRUM
 * @input_type Matrix<complex<float> >
 * @input_description Windowed spectrum data. A row index is channel, and a column index is frequency;
 *
 * @input_name SOURCES
 * @input_type Vector<ObjectRef>
 * @input_description Source locations with ID. Each element of the vector is a source location with ID specified by "Source".
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description Source separation results. Each element of the vector is a spectrum of separated sound.
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The shift length between adjacent frames (in samples)[default: 160].
 *
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description Sampling rate (Hz)[default: 16000].
 *
 * @parameter_name RANGE_MIN
 * @parameter_type float
 * @parameter_value 5.0
 * @parameter_description Separation range at direction of 0 degree (degree)[default: 5.0].
 *
 * @parameter_name RANGE_MAX
 * @parameter_type float
 * @parameter_value 30.0
 * @parameter_description Separation range at direction of 90 degree (degree)[default: 30.0].
 *
 * @parameter_name SEPARATE_METHOD
 * @parameter_type string
 * @parameter_value EPIPOLAR
 * @parameter_list EPIPOLAR:HRTFMESURED
 * @parameter_description Method of localization [default: EPIPOLAR].
 *
 * @parameter_name HARK_DAT
 * @parameter_type string
 * @parameter_value hark.dat
 * @parameter_description Transfer function file.
 *
 END*/


class SeparateBinaural: public BufferedNode
{
    int spectrumInputID, sourceInputID;
    int outputID;
    int samplingRate;

    SeparationSetting::ptr_t separationSetting;

    bool prevFindID1; // debug

    float rangeMin, rangeMax;

    BGNEstimator bgnEstimator;
    shared_ptr<IIPDS> iipds;
    String separationMethod;

    Logger& logger;

public:
    SeparateBinaural(string nodeName, ParameterSet params) :
        BufferedNode(nodeName, params),
        logger(Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME))
    {

        /*
         * 入出力端子の設定
         */
        spectrumInputID = addInput("SPECTRUM");
        sourceInputID = addInput("SOURCES");
        outputID = addOutput("OUTPUT");

        samplingRate = getParameter<int>("SAMPLING_RATE");

        prevFindID1 = false;

        rangeMin = getParameter<float>("RANGE_MIN");
        rangeMax = getParameter<float>("RANGE_MAX");

        separationMethod = object_cast<String>(parameters.get("SEPARATE_METHOD"));

        inOrder = true;
    }

    virtual ~SeparateBinaural()
    {
    }

    /**
     * ノードのパラメータ取得メソッド
     */
    template<typename T>
    T getParameter(string name)
    {
        return dereference_cast<T> (parameters.get(name));
    }

    void count0Initialize(int output_id, Buffer &out)
    {
        // 入力を取得する
        RCPtr<Matrix<complex<float> > > spectrumInputPtr = getInput(this->spectrumInputID, 0);
        Matrix<complex<float> >& spectrumInput = *spectrumInputPtr;
        //RCPtr<Vector<RCPtr<Source> > > sourceInputPtr = getInput(this->sourceInputID, count);
        //Vector<RCPtr<Source> >& sourceInput = *sourceInputPtr;

        // 音源分離の設定
        this->separationSetting = SeparationSetting::ptr_t(new SeparationSetting());
        this->separationSetting->ndft = (spectrumInput.ncols() - 1) * 2;
        this->separationSetting->samplingFrequency = Frequency::FromHz(this->samplingRate);

        harkio::HARK hark;
        //hark.readFile("/home/harkuser/myaudition/hark.dat");
        hark.readFile(object_cast<String>(parameters.get("HARK_DAT")));

        // LocalizeBinauralだけチャンネル選択に対応した。
        // こちらはとりあえずLeft == 0, Right == 1にしている。
        // こちらもチャンネル選択に対応する場合はLocalizeBinauralを参考に改造しよう。
        iipds = HRTFToIIPDS(hark, 0, 1);
    }

    void calculate(int output_id, int count, Buffer &out)
    {
        LogFine(logger, "inputting variables");
        //DebugPuts("inputting variables");

        // 入力を取得する
        RCPtr<Matrix<complex<float> > > spectrumInputPtr = getInput(this->spectrumInputID, count);
        Matrix<complex<float> >& spectrumInput = *spectrumInputPtr;
        //RCPtr<Vector<RCPtr<Source> > > sourceInputPtr = getInput(this->sourceInputID, count);
        //Vector<RCPtr<Source> >& sourceInput = *sourceInputPtr;
        RCPtr<Vector<ObjectRef> > sourceInputPtr = getInput(this->sourceInputID, count);
        Vector<ObjectRef>& sourceInput = *sourceInputPtr;

        LogFine(logger, "inspecting variables");
        //DebugPuts("inspecting variables");

        // 入力を検査する
        if (spectrumInput.nrows() != 2)
        {
            throw new InvalidInputException(
                (format("Input matrix is not 2ch, but %dch") % spectrumInput.nrows()).str());
        }

        if (count == 0)
        {
            // 最初のカウントのみ初期化する
            count0Initialize(output_id, out);
        }

        // 入力を扱いやすい形へ変換する
        shared_ptr<vector<Spectrum::ptr_t> > spectrumArray = ToSignalArray<Spectrum::ElementType>(spectrumInput);
        LRSpectra::ptr_t lrspectra(new LRSpectra(spectrumArray->at(0), spectrumArray->at(1)));

        // 両耳間強度差と両耳間位相差を計算し生成する
        IID::ptr_t iid(new IID(lrspectra, Channel::R));
        IPD::ptr_t ipd(new IPD(lrspectra, Channel::R));

        LRUnspectra::ptr_t lrbgnoise;
        // 内蔵の推定器でBGNを推定する
        this->bgnEstimator.Estimate(lrspectra);
        if (this->bgnEstimator.Estimated())
        {
            lrbgnoise = this->bgnEstimator.BGNoise();
        }


        // 出力用マップの生成
        //RCPtr<Map<int, ObjectRef> > outputMapPtr = ObjectRef(new Map<int, ObjectRef>);
        //Map<int,ObjectRef> &outputMap = object_cast<Map<int, ObjectRef> >(outputMapPtr);
        RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
        Map<int,ObjectRef> &outputMap = *outputMapPtr;
        outputMap.clear() ;

        if (lrbgnoise)
        {
            // 音源分離器の生成
            Separation::ptr_t separation;

            if (separationMethod == "HRTFMESURED")
            {
                separation = Separation::ptr_t(new HRTFSeparation(lrspectra, this->separationSetting, lrbgnoise, *iipds, 10));
            }
            else
            {
                separation = Separation::ptr_t(new EpipolarSeparation(lrspectra, this->separationSetting));
            }


            RCPtr<Source> source;

            LogFine(logger, format("got %d sources") % sourceInput.size());
            //DebugPrintF(format("got %d sources") % sourceInput.size());

            bool findID1 = false;

            Angle rangeMin_ = Angle::FromDegree(rangeMin);
            Angle rangeMax_ = Angle::FromDegree(rangeMax);

            // 分離する
            for (size_t i = 0; i < sourceInput.size(); ++i)
            {
                source = sourceInput[i];
                float x = source->x[0];
                float y = source->x[1];
                int id = source->id;
                if (id == 1)
                {
                    findID1 = true;
                }

                // 分離する方向と範囲を計算
                Angle dir(atan2(y, x));
                Angle range(getPermissibleDirectionRange(dir, rangeMin_, rangeMax_));

                // デバッグ表示
                LogFine(logger, format("%10d source id = %d, x = %f, y = %f , angle = %f deg, range = %f deg")
                        % count % id % x % y % dir.ToDegree() % range.ToDegree());

                // 方向と範囲を用いて分離
                Separation::LRSpectra::ptr_t separatedSpectra =
                    separation->separate(dir, range);

                // ミックスしてモノラルにするのではなく、
                // 右の音は右チャンネルだけ、左の音は左チャンネルだけ
                // にしたほうが性能が上がるらしい
                //Spectrum::ptr_t monaural = MixToMonaural(separatedSpectra);

                // とりあえず固定的に左を出力
                Spectrum::ptr_t monaural;
                if (dir < Angle::FromDegree(0))
                {
                    monaural = separatedSpectra->Left();
                }
                else
                {
                    monaural = separatedSpectra->Right();
                }

                // 分離した信号を出力用に変換
                FD::RCPtr<FD::Vector<complex<float> > > outVector = ToFDVector(monaural);

                // 出力用マップに追加
                outputMap.insert(make_pair(id, outVector));
            }

            if (prevFindID1 == true && findID1 == false)
            {
                LogFine(logger, format("find id 1 drop out"));
            }
            prevFindID1 = findID1;
        }

        // 出力
        (*outputs[outputID].buffer)[count] = outputMapPtr;
    }

};

